@@ -5,9 +5,8 @@ import (
 	"context"
 	"crypto/tls"
 	"crypto/x509"
-	"errors"
 	"fmt"
-	"io/ioutil"
+	"io/fs"
 	"net"
 	"net/http"
 	"net/http/pprof"
@@ -16,22 +15,22 @@ import (
 	"os/signal"
 	"path/filepath"
 	"runtime"
-	"strconv"
 	"sync"
 	"syscall"
 	"time"
 
-	"github.com/AdguardTeam/AdGuardHome/internal/agherr"
 	"github.com/AdguardTeam/AdGuardHome/internal/aghnet"
 	"github.com/AdguardTeam/AdGuardHome/internal/aghos"
 	"github.com/AdguardTeam/AdGuardHome/internal/dhcpd"
-	"github.com/AdguardTeam/AdGuardHome/internal/dnsfilter"
 	"github.com/AdguardTeam/AdGuardHome/internal/dnsforward"
+	"github.com/AdguardTeam/AdGuardHome/internal/filtering"
 	"github.com/AdguardTeam/AdGuardHome/internal/querylog"
 	"github.com/AdguardTeam/AdGuardHome/internal/stats"
 	"github.com/AdguardTeam/AdGuardHome/internal/updater"
 	"github.com/AdguardTeam/AdGuardHome/internal/version"
+	"github.com/AdguardTeam/golibs/errors"
 	"github.com/AdguardTeam/golibs/log"
+	"github.com/AdguardTeam/golibs/netutil"
 	"gopkg.in/natefinch/lumberjack.v2"
 )
 
@@ -45,20 +44,25 @@ type homeContext struct {
 	// Modules
 	// --
 
-	clients    clientsContainer          // per-client-settings module
-	stats      stats.Stats               // statistics module
-	queryLog   querylog.QueryLog         // query log module
-	dnsServer  *dnsforward.Server        // DNS module
-	rdns       *RDNS                     // rDNS module
-	whois      *Whois                    // WHOIS module
-	dnsFilter  *dnsfilter.DNSFilter      // DNS filtering module
-	dhcpServer *dhcpd.Server             // DHCP module
-	auth       *Auth                     // HTTP authentication module
-	filters    Filtering                 // DNS filtering module
-	web        *Web                      // Web (HTTP, HTTPS) module
-	tls        *TLSMod                   // TLS module
-	etcHosts   *aghnet.EtcHostsContainer // IP-hostname pairs taken from system configuration (e.g. /etc/hosts) files
-	updater    *updater.Updater
+	clients    clientsContainer     // per-client-settings module
+	stats      stats.Stats          // statistics module
+	queryLog   querylog.QueryLog    // query log module
+	dnsServer  *dnsforward.Server   // DNS module
+	rdns       *RDNS                // rDNS module
+	whois      *WHOIS               // WHOIS module
+	dnsFilter  *filtering.DNSFilter // DNS filtering module
+	dhcpServer *dhcpd.Server        // DHCP module
+	auth       *Auth                // HTTP authentication module
+	filters    Filtering            // DNS filtering module
+	web        *Web                 // Web (HTTP, HTTPS) module
+	tls        *TLSMod              // TLS module
+	// etcHosts is an IP-hostname pairs set taken from system configuration
+	// (e.g. /etc/hosts) files.
+	etcHosts *aghnet.HostsContainer
+	// hostsWatcher is the watcher to detect changes in the hosts files.
+	hostsWatcher aghos.FSWatcher
+
+	updater *updater.Updater
 
 	subnetDetector *aghnet.SubnetDetector
 
@@ -70,7 +74,7 @@ type homeContext struct {
 
 	configFilename   string // Config filename (can be overridden via the command line arguments)
 	workDir          string // Location of our directory, used to protect against CWD being somewhere else
-	firstRun         bool   // if set to true, don't run any services except HTTP web inteface, and serve only first-run html
+	firstRun         bool   // if set to true, don't run any services except HTTP web interface, and serve only first-run html
 	pidFileName      string // PID file name.  Empty if no PID file was created.
 	disableUpdate    bool   // If set, don't check for updates
 	controlLock      sync.Mutex
@@ -92,7 +96,7 @@ func (c *homeContext) getDataDir() string {
 var Context homeContext
 
 // Main is the entry point
-func Main() {
+func Main(clientBuildFS fs.FS) {
 	// config can be specified, which reads options from there, but other command line flags have to override config values
 	// therefore, we must do it manually instead of using a lib
 	args := loadOptions()
@@ -117,12 +121,13 @@ func Main() {
 	}()
 
 	if args.serviceControlAction != "" {
-		handleServiceControlAction(args)
+		handleServiceControlAction(args, clientBuildFS)
+
 		return
 	}
 
 	// run the protection
-	run(args)
+	run(args, clientBuildFS)
 }
 
 func setupContext(args options) {
@@ -154,14 +159,11 @@ func setupContext(args options) {
 	}
 
 	if !Context.firstRun {
-		// Do the upgrade if necessary
+		// Do the upgrade if necessary.
 		err := upgradeConfig()
-		if err != nil {
-			log.Fatal(err)
-		}
+		fatalOnError(err)
 
-		err = parseConfig()
-		if err != nil {
+		if err = parseConfig(); err != nil {
 			log.Error("parsing configuration file: %s", err)
 
 			os.Exit(1)
@@ -177,18 +179,99 @@ func setupContext(args options) {
 	Context.mux = http.NewServeMux()
 }
 
-func setupConfig(args options) {
+// logIfUnsupported logs a formatted warning if the error is one of the
+// unsupported errors and returns nil.  If err is nil, logIfUnsupported returns
+// nil.  Otherise, it returns err.
+func logIfUnsupported(msg string, err error) (outErr error) {
+	if errors.As(err, new(*aghos.UnsupportedError)) {
+		log.Debug(msg, err)
+
+		return nil
+	}
+
+	return err
+}
+
+// configureOS sets the OS-related configuration.
+func configureOS(conf *configuration) (err error) {
+	osConf := conf.OSConfig
+	if osConf == nil {
+		return nil
+	}
+
+	if osConf.Group != "" {
+		err = aghos.SetGroup(osConf.Group)
+		err = logIfUnsupported("warning: setting group", err)
+		if err != nil {
+			return fmt.Errorf("setting group: %w", err)
+		}
+
+		log.Info("group set to %s", osConf.Group)
+	}
+
+	if osConf.User != "" {
+		err = aghos.SetUser(osConf.User)
+		err = logIfUnsupported("warning: setting user", err)
+		if err != nil {
+			return fmt.Errorf("setting user: %w", err)
+		}
+
+		log.Info("user set to %s", osConf.User)
+	}
+
+	if osConf.RlimitNoFile != 0 {
+		err = aghos.SetRlimit(osConf.RlimitNoFile)
+		err = logIfUnsupported("warning: setting rlimit", err)
+		if err != nil {
+			return fmt.Errorf("setting rlimit: %w", err)
+		}
+
+		log.Info("rlimit_nofile set to %d", osConf.RlimitNoFile)
+	}
+
+	return nil
+}
+
+// setupHostsContainer initializes the structures to keep up-to-date the hosts
+// provided by the OS.
+func setupHostsContainer() (err error) {
+	Context.hostsWatcher, err = aghos.NewOSWritesWatcher()
+	if err != nil {
+		return fmt.Errorf("initing hosts watcher: %w", err)
+	}
+
+	Context.etcHosts, err = aghnet.NewHostsContainer(
+		filtering.SysHostsListID,
+		aghos.RootDirFS(),
+		Context.hostsWatcher,
+		aghnet.DefaultHostsPaths()...,
+	)
+	if err != nil {
+		cerr := Context.hostsWatcher.Close()
+		if errors.Is(err, aghnet.ErrNoHostsPaths) && cerr == nil {
+			log.Info("warning: initing hosts container: %s", err)
+
+			return nil
+		}
+
+		return errors.WithDeferred(fmt.Errorf("initing hosts container: %w", err), cerr)
+	}
+
+	return nil
+}
+
+func setupConfig(args options) (err error) {
 	config.DHCP.WorkDir = Context.workDir
 	config.DHCP.HTTPRegister = httpRegister
 	config.DHCP.ConfigModified = onConfigModified
 
-	Context.dhcpServer = dhcpd.Create(config.DHCP)
-	if Context.dhcpServer == nil {
+	Context.dhcpServer, err = dhcpd.Create(config.DHCP)
+	if Context.dhcpServer == nil || err != nil {
 		// TODO(a.garipov): There are a lot of places in the code right
 		// now which assume that the DHCP server can be nil despite this
 		// condition.  Inspect them and perhaps rewrite them to use
 		// Enabled() instead.
-		log.Fatalf("can't initialize dhcp module")
+		return fmt.Errorf("initing dhcp: %w", err)
 	}
 
 	Context.updater = updater.NewUpdater(&updater.Config{
@@ -204,31 +287,97 @@ func setupConfig(args options) {
 	})
 
 	if !args.noEtcHosts {
-		Context.etcHosts = &aghnet.EtcHostsContainer{}
-		Context.etcHosts.Init("")
+		if err = setupHostsContainer(); err != nil {
+			return err
+		}
 	}
-	Context.clients.Init(config.Clients, Context.dhcpServer, Context.etcHosts)
-	config.Clients = nil
 
-	if (runtime.GOOS == "linux" || runtime.GOOS == "darwin") &&
-		config.RlimitNoFile != 0 {
-		aghos.SetRlimit(config.RlimitNoFile)
+	Context.clients.Init(config.Clients, Context.dhcpServer, Context.etcHosts)
+
+	if args.bindPort != 0 {
+		pm := portsMap{}
+		pm.add(
+			args.bindPort,
+			config.BetaBindPort,
+			config.DNS.Port,
+		)
+		if config.TLS.Enabled {
+			pm.add(
+				config.TLS.PortHTTPS,
+				config.TLS.PortDNSOverTLS,
+				config.TLS.PortDNSOverQUIC,
+				config.TLS.PortDNSCrypt,
+			)
+		}
+		if err = pm.validate(); err != nil {
+			return err
+		}
+
+		config.BindPort = args.bindPort
 	}
 
 	// override bind host/port from the console
 	if args.bindHost != nil {
 		config.BindHost = args.bindHost
 	}
-	if args.bindPort != 0 {
-		config.BindPort = args.bindPort
-	}
 	if len(args.pidFile) != 0 && writePIDFile(args.pidFile) {
 		Context.pidFileName = args.pidFile
 	}
+
+	return nil
+}
+
+func initWeb(args options, clientBuildFS fs.FS) (web *Web, err error) {
+	var clientFS, clientBetaFS fs.FS
+	if args.localFrontend {
+		log.Info("warning: using local frontend files")
+
+		clientFS = os.DirFS("build/static")
+		clientBetaFS = os.DirFS("build2/static")
+	} else {
+		clientFS, err = fs.Sub(clientBuildFS, "build/static")
+		if err != nil {
+			return nil, fmt.Errorf("getting embedded client subdir: %w", err)
+		}
+
+		clientBetaFS, err = fs.Sub(clientBuildFS, "build2/static")
+		if err != nil {
+			return nil, fmt.Errorf("getting embedded beta client subdir: %w", err)
+		}
+	}
+
+	webConf := webConfig{
+		firstRun:     Context.firstRun,
+		BindHost:     config.BindHost,
+		BindPort:     config.BindPort,
+		BetaBindPort: config.BetaBindPort,
+
+		ReadTimeout:       readTimeout,
+		ReadHeaderTimeout: readHdrTimeout,
+		WriteTimeout:      writeTimeout,
+
+		clientFS:     clientFS,
+		clientBetaFS: clientBetaFS,
+	}
+
+	web = CreateWeb(&webConf)
+	if web == nil {
+		return nil, fmt.Errorf("initializing web: %w", err)
+	}
+
+	return web, nil
+}
+
+func fatalOnError(err error) {
+	if err != nil {
+		log.Fatal(err)
+	}
 }
 
 // run performs configurating and starts AdGuard Home.
-func run(args options) {
+func run(args options, clientBuildFS fs.FS) {
+	var err error
+
 	// configure config filename
 	initConfigFilename(args)
 
@@ -250,19 +399,21 @@ func run(args options) {
 
 	setupContext(args)
 
-	// clients package uses dnsfilter package's static data (dnsfilter.BlockedSvcKnown()),
-	//  so we have to initialize dnsfilter's static data first,
-	//  but also avoid relying on automatic Go init() function
-	dnsfilter.InitModule()
+	err = configureOS(config)
+	fatalOnError(err)
 
-	setupConfig(args)
+	// clients package uses filtering package's static data (filtering.BlockedSvcKnown()),
+	//  so we have to initialize filtering's static data first,
+	//  but also avoid relying on automatic Go init() function
+	filtering.InitModule()
+
+	err = setupConfig(args)
+	fatalOnError(err)
 
 	if !Context.firstRun {
 		// Save the updated config
-		err := config.write()
-		if err != nil {
-			log.Fatal(err)
-		}
+		err = config.write()
+		fatalOnError(err)
 
 		if config.DebugPProf {
 			mux := http.NewServeMux()
@@ -279,7 +430,7 @@ func run(args options) {
 		}
 	}
 
-	err := os.MkdirAll(Context.getDataDir(), 0o755)
+	err = os.MkdirAll(Context.getDataDir(), 0o755)
 	if err != nil {
 		log.Fatalf("Cannot create DNS data dir at %s: %s", Context.getDataDir(), err)
 	}
@@ -295,6 +446,7 @@ func run(args options) {
 	} else {
 		log.Info("authratelimiter is disabled")
 	}
+
 	Context.auth = InitAuth(
 		sessFilename,
 		config.Users,
@@ -311,40 +463,23 @@ func run(args options) {
 		log.Fatalf("Can't initialize TLS module")
 	}
 
-	webConf := webConfig{
-		firstRun:     Context.firstRun,
-		BindHost:     config.BindHost,
-		BindPort:     config.BindPort,
-		BetaBindPort: config.BetaBindPort,
-
-		ReadTimeout:       readTimeout,
-		ReadHeaderTimeout: readHdrTimeout,
-		WriteTimeout:      writeTimeout,
-	}
-	Context.web = CreateWeb(&webConf)
-	if Context.web == nil {
-		log.Fatalf("Can't initialize Web module")
-	}
+	Context.web, err = initWeb(args, clientBuildFS)
+	fatalOnError(err)
 
 	Context.subnetDetector, err = aghnet.NewSubnetDetector()
-	if err != nil {
-		log.Fatal(err)
-	}
+	fatalOnError(err)
 
 	if !Context.firstRun {
 		err = initDNSServer()
-		if err != nil {
-			log.Fatalf("%s", err)
-		}
+		fatalOnError(err)
 
 		Context.tls.Start()
-		Context.etcHosts.Start()
 
 		go func() {
 			serr := startDNSServer()
 			if serr != nil {
 				closeDNSServer()
-				log.Fatal(serr)
+				fatalOnError(serr)
 			}
 		}()
 
@@ -411,7 +546,7 @@ Please note, that this is crucial for a server to be able to use privileged port
 You have two options:
 1. Run AdGuard Home with root privileges
 2. On Linux you can grant the CAP_NET_BIND_SERVICE capability:
-https://github.com/AdguardTeam/AdGuardHome/internal/wiki/Getting-Started#running-without-superuser`
+https://github.com/AdguardTeam/AdGuardHome/wiki/Getting-Started#running-without-superuser`
 
 		log.Fatal(msg)
 	}
@@ -426,7 +561,7 @@ Please note, that this is crucial for a DNS server to be able to use that port.`
 // Write PID to a file
 func writePIDFile(fn string) bool {
 	data := fmt.Sprintf("%d", os.Getpid())
-	err := ioutil.WriteFile(fn, []byte(data), 0o644)
+	err := os.WriteFile(fn, []byte(data), 0o644)
 	if err != nil {
 		log.Error("Couldn't write PID to file %s: %v", fn, err)
 		return false
@@ -538,7 +673,7 @@ func configureLogger(args options) {
 
 // cleanup stops and resets all the modules.
 func cleanup(ctx context.Context) {
-	log.Info("Stopping AdGuard Home")
+	log.Info("stopping AdGuard Home")
 
 	if Context.web != nil {
 		Context.web.Close(ctx)
@@ -551,14 +686,28 @@ func cleanup(ctx context.Context) {
 
 	err := stopDNSServer()
 	if err != nil {
-		log.Error("Couldn't stop DNS server: %s", err)
+		log.Error("stopping dns server: %s", err)
 	}
 
 	if Context.dhcpServer != nil {
-		Context.dhcpServer.Stop()
+		err = Context.dhcpServer.Stop()
+		if err != nil {
+			log.Error("stopping dhcp server: %s", err)
+		}
 	}
 
-	Context.etcHosts.Close()
+	if Context.etcHosts != nil {
+		// Currently Context.hostsWatcher is only used in Context.etcHosts and
+		// needs closing only in case of the successful initialization of
+		// Context.etcHosts.
+		if err = Context.hostsWatcher.Close(); err != nil {
+			log.Error("closing hosts watcher: %s", err)
+		}
+
+		if err = Context.etcHosts.Close(); err != nil {
+			log.Error("closing hosts container: %s", err)
+		}
+	}
 
 	if Context.tls != nil {
 		Context.tls.Close()
@@ -599,7 +748,28 @@ func loadOptions() options {
 	return o
 }
 
-// printHTTPAddresses prints the IP addresses which user can use to open the
+// printWebAddrs prints addresses built from proto, addr, and an appropriate
+// port.  At least one address is printed with the value of port.  If the value
+// of betaPort is 0, the second address is not printed.  The output example:
+//
+//   Go to http://127.0.0.1:80
+//   Go to http://127.0.0.1:3000 (BETA)
+//
+func printWebAddrs(proto, addr string, port, betaPort int) {
+	const (
+		hostMsg     = "Go to %s://%s"
+		hostBetaMsg = hostMsg + " (BETA)"
+	)
+
+	log.Printf(hostMsg, proto, netutil.JoinHostPort(addr, port))
+	if betaPort == 0 {
+		return
+	}
+
+	log.Printf(hostBetaMsg, proto, netutil.JoinHostPort(addr, config.BetaBindPort))
+}
+
+// printHTTPAddresses prints the IP addresses which user can use to access the
 // admin interface.  proto is either schemeHTTP or schemeHTTPS.
 func printHTTPAddresses(proto string) {
 	tlsConf := tlsConfigSettings{}
@@ -607,45 +777,39 @@ func printHTTPAddresses(proto string) {
 		Context.tls.WriteDiskConfig(&tlsConf)
 	}
 
-	port := strconv.Itoa(config.BindPort)
+	port := config.BindPort
 	if proto == schemeHTTPS {
-		port = strconv.Itoa(tlsConf.PortHTTPS)
+		port = tlsConf.PortHTTPS
 	}
 
-	var hostStr string
+	// TODO(e.burkov): Inspect and perhaps merge with the previous condition.
 	if proto == schemeHTTPS && tlsConf.ServerName != "" {
-		if tlsConf.PortHTTPS == 443 {
-			log.Printf("Go to https://%s", tlsConf.ServerName)
-		} else {
-			log.Printf("Go to https://%s:%s", tlsConf.ServerName, port)
-		}
-	} else if config.BindHost.IsUnspecified() {
-		log.Println("AdGuard Home is available on the following addresses:")
-		ifaces, err := aghnet.GetValidNetInterfacesForWeb()
-		if err != nil {
-			// That's weird, but we'll ignore it
-			hostStr = config.BindHost.String()
-			log.Printf("Go to %s://%s", proto, net.JoinHostPort(hostStr, port))
-			if config.BetaBindPort != 0 {
-				log.Printf("Go to %s://%s (BETA)", proto, net.JoinHostPort(hostStr, strconv.Itoa(config.BetaBindPort)))
-			}
-			return
-		}
+		printWebAddrs(proto, tlsConf.ServerName, tlsConf.PortHTTPS, 0)
 
-		for _, iface := range ifaces {
-			for _, addr := range iface.Addresses {
-				hostStr = addr.String()
-				log.Printf("Go to %s://%s", proto, net.JoinHostPort(hostStr, strconv.Itoa(config.BindPort)))
-				if config.BetaBindPort != 0 {
-					log.Printf("Go to %s://%s (BETA)", proto, net.JoinHostPort(hostStr, strconv.Itoa(config.BetaBindPort)))
-				}
-			}
-		}
-	} else {
-		hostStr = config.BindHost.String()
-		log.Printf("Go to %s://%s", proto, net.JoinHostPort(hostStr, port))
-		if config.BetaBindPort != 0 {
-			log.Printf("Go to %s://%s (BETA)", proto, net.JoinHostPort(hostStr, strconv.Itoa(config.BetaBindPort)))
+		return
+	}
+
+	bindhost := config.BindHost
+	if !bindhost.IsUnspecified() {
+		printWebAddrs(proto, bindhost.String(), port, config.BetaBindPort)
+
+		return
+	}
+
+	ifaces, err := aghnet.GetValidNetInterfacesForWeb()
+	if err != nil {
+		log.Error("web: getting iface ips: %s", err)
+		// That's weird, but we'll ignore it.
+		//
+		// TODO(e.burkov): Find out when it happens.
+		printWebAddrs(proto, bindhost.String(), port, config.BetaBindPort)
+
+		return
+	}
+
+	for _, iface := range ifaces {
+		for _, addr := range iface.Addresses {
+			printWebAddrs(proto, addr.String(), config.BindPort, config.BetaBindPort)
 		}
 	}
 }
@@ -704,7 +868,7 @@ func customDialContext(ctx context.Context, network, addr string) (conn net.Conn
 		return conn, err
 	}
 
-	return nil, agherr.Many(fmt.Sprintf("couldn't dial to %s", addr), dialErrs...)
+	return nil, errors.List(fmt.Sprintf("couldn't dial to %s", addr), dialErrs...)
 }
 
 func getHTTPProxy(_ *http.Request) (*url.URL, error) {
