@@ -2,20 +2,23 @@ package home
 
 import (
 	"fmt"
-	"io/ioutil"
+	"io/fs"
 	"os"
 	"runtime"
 	"strconv"
 	"strings"
 	"syscall"
+	"time"
 
 	"github.com/AdguardTeam/AdGuardHome/internal/aghos"
+	"github.com/AdguardTeam/AdGuardHome/internal/version"
+	"github.com/AdguardTeam/golibs/errors"
 	"github.com/AdguardTeam/golibs/log"
 	"github.com/kardianos/service"
 )
 
-// TODO(a.garipov): Move shell templates into actual files.  Either during the
-// v0.106.0 cycle using packr or during the following cycle using go:embed.
+// TODO(a.garipov): Consider moving the shell templates into actual files and
+// using go:embed instead of using large string constants.
 
 const (
 	launchdStdoutPath  = "/var/log/AdGuardHome.stdout.log"
@@ -25,31 +28,37 @@ const (
 	serviceDescription = "AdGuard Home: Network-level blocker"
 )
 
-// Represents the program that will be launched by a service or daemon
+// program represents the program that will be launched by as a service or a
+// daemon.
 type program struct {
-	opts options
+	clientBuildFS fs.FS
+	opts          options
 }
 
-// Start should quickly start the program
-func (p *program) Start(s service.Service) error {
-	// Start should not block. Do the actual work async.
+// Start implements service.Interface interface for *program.
+func (p *program) Start(_ service.Service) (err error) {
+	// Start should not block.  Do the actual work async.
 	args := p.opts
 	args.runningAsService = true
-	go run(args)
+
+	go run(args, p.clientBuildFS)
+
 	return nil
 }
 
-// Stop stops the program
-func (p *program) Stop(s service.Service) error {
-	// Stop should not block. Return with a few seconds.
+// Stop implements service.Interface interface for *program.
+func (p *program) Stop(_ service.Service) error {
+	// Stop should not block.  Return with a few seconds.
 	if Context.appSignalChannel == nil {
 		os.Exit(0)
 	}
+
 	Context.appSignalChannel <- syscall.SIGINT
+
 	return nil
 }
 
-// svcStatus check the service's status.
+// svcStatus returns the service's status.
 //
 // On OpenWrt, the service utility may not exist.  We use our service script
 // directly in this case.
@@ -84,48 +93,57 @@ func svcAction(s service.Service, action string) (err error) {
 	return err
 }
 
-// Send SIGHUP to a process with ID taken from our pid-file
-// If pid-file doesn't exist, find our PID using 'ps' command
+// Send SIGHUP to a process with PID taken from our .pid file.  If it doesn't
+// exist, find our PID using 'ps' command.
 func sendSigReload() {
 	if runtime.GOOS == "windows" {
-		log.Error("Not implemented on Windows")
+		log.Error("service: not implemented on windows")
+
 		return
 	}
 
 	pidfile := fmt.Sprintf("/var/run/%s.pid", serviceName)
-	data, err := ioutil.ReadFile(pidfile)
-	if os.IsNotExist(err) {
-		var code int
-		var psdata string
-		code, psdata, err = aghos.RunCommand("ps", "-C", serviceName, "-o", "pid=")
-		if err != nil || code != 0 {
-			log.Error("Can't find AdGuardHome process: %s  code:%d", err, code)
+	var pid int
+	data, err := os.ReadFile(pidfile)
+	if errors.Is(err, os.ErrNotExist) {
+		if pid, err = aghos.PIDByCommand(serviceName, os.Getpid()); err != nil {
+			log.Error("service: finding AdGuardHome process: %s", err)
+
 			return
 		}
-		data = []byte(psdata)
-
 	} else if err != nil {
-		log.Error("Can't read PID file %s: %s", pidfile, err)
+		log.Error("service: reading pid file %s: %s", pidfile, err)
+
+		return
+	} else {
+		parts := strings.SplitN(string(data), "\n", 2)
+		if len(parts) == 0 {
+			log.Error("service: parsing pid file %s: bad value", pidfile)
+
+			return
+		}
+
+		if pid, err = strconv.Atoi(strings.TrimSpace(parts[0])); err != nil {
+			log.Error("service: parsing pid from file %s: %s", pidfile, err)
+
+			return
+		}
+	}
+
+	var proc *os.Process
+	if proc, err = os.FindProcess(pid); err != nil {
+		log.Error("service: finding process for pid %d: %s", pid, err)
+
 		return
 	}
 
-	parts := strings.SplitN(string(data), "\n", 2)
-	if len(parts) == 0 {
-		log.Error("Can't read PID file %s: bad value", pidfile)
+	if err = proc.Signal(syscall.SIGHUP); err != nil {
+		log.Error("service: sending signal HUP to pid %d: %s", pid, err)
+
 		return
 	}
 
-	pid, err := strconv.Atoi(strings.TrimSpace(parts[0]))
-	if err != nil {
-		log.Error("Can't read PID file %s: %s", pidfile, err)
-		return
-	}
-	err = aghos.SendProcessSignal(pid, syscall.SIGHUP)
-	if err != nil {
-		log.Error("Can't send signal to PID %d: %s", pid, err)
-		return
-	}
-	log.Debug("Sent signal to PID %d", pid)
+	log.Debug("service: sent signal to pid %d", pid)
 }
 
 // handleServiceControlAction one of the possible control actions:
@@ -138,19 +156,25 @@ func sendSigReload() {
 // run - this is a special command that is not supposed to be used directly
 // it is specified when we register a service, and it indicates to the app
 // that it is being run as a service/daemon.
-func handleServiceControlAction(opts options) {
+func handleServiceControlAction(opts options, clientBuildFS fs.FS) {
+	// Call chooseSystem expicitly to introduce OpenBSD support for service
+	// package.  It's a noop for other GOOS values.
+	chooseSystem()
+
 	action := opts.serviceControlAction
-	log.Printf("Service control action: %s", action)
+	log.Printf("service: control action: %s", action)
 
 	if action == "reload" {
 		sendSigReload()
+
 		return
 	}
 
 	pwd, err := os.Getwd()
 	if err != nil {
-		log.Fatal("Unable to find the path to the current directory")
+		log.Fatalf("service: getting current directory: %s", err)
 	}
+
 	runOpts := opts
 	runOpts.serviceControlAction = "run"
 	svcConfig := &service.Config{
@@ -161,49 +185,52 @@ func handleServiceControlAction(opts options) {
 		Arguments:        serialize(runOpts),
 	}
 	configureService(svcConfig)
-	prg := &program{runOpts}
-	s, err := service.New(prg, svcConfig)
-	if err != nil {
-		log.Fatal(err)
+
+	prg := &program{
+		clientBuildFS: clientBuildFS,
+		opts:          runOpts,
+	}
+	var s service.Service
+	if s, err = service.New(prg, svcConfig); err != nil {
+		log.Fatalf("service: initializing service: %s", err)
 	}
 
-	if action == "status" {
+	switch action {
+	case "status":
 		handleServiceStatusCommand(s)
-	} else if action == "run" {
-		err = s.Run()
-		if err != nil {
-			log.Fatalf("Failed to run service: %s", err)
+	case "run":
+		if err = s.Run(); err != nil {
+			log.Fatalf("service: failed to run service: %s", err)
 		}
-	} else if action == "install" {
+	case "install":
 		initConfigFilename(opts)
 		initWorkingDir(opts)
 		handleServiceInstallCommand(s)
-	} else if action == "uninstall" {
+	case "uninstall":
 		handleServiceUninstallCommand(s)
-	} else {
-		err = svcAction(s, action)
-		if err != nil {
-			log.Fatal(err)
+	default:
+		if err = svcAction(s, action); err != nil {
+			log.Fatalf("service: executing action %q: %s", action, err)
 		}
 	}
 
-	log.Printf("Action %s has been done successfully on %s", action, service.ChosenSystem().String())
+	log.Printf("service: action %s has been done successfully on %s", action, service.ChosenSystem())
 }
 
-// handleServiceStatusCommand handles service "status" command
+// handleServiceStatusCommand handles service "status" command.
 func handleServiceStatusCommand(s service.Service) {
 	status, errSt := svcStatus(s)
 	if errSt != nil {
-		log.Fatalf("failed to get service status: %s", errSt)
+		log.Fatalf("service: failed to get service status: %s", errSt)
 	}
 
 	switch status {
 	case service.StatusUnknown:
-		log.Printf("Service status is unknown")
+		log.Printf("service: status is unknown")
 	case service.StatusStopped:
-		log.Printf("Service is stopped")
+		log.Printf("service: stopped")
 	case service.StatusRunning:
-		log.Printf("Service is running")
+		log.Printf("service: running")
 	}
 }
 
@@ -211,30 +238,32 @@ func handleServiceStatusCommand(s service.Service) {
 func handleServiceInstallCommand(s service.Service) {
 	err := svcAction(s, "install")
 	if err != nil {
-		log.Fatal(err)
+		log.Fatalf("service: executing action %q: %s", "install", err)
 	}
 
 	if aghos.IsOpenWrt() {
-		// On OpenWrt it is important to run enable after the service installation
-		// Otherwise, the service won't start on the system startup
+		// On OpenWrt it is important to run enable after the service
+		// installation Otherwise, the service won't start on the system
+		// startup.
 		_, err = runInitdCommand("enable")
 		if err != nil {
-			log.Fatal(err)
+			log.Fatalf("service: running init enable: %s", err)
 		}
 	}
 
-	// Start automatically after install
+	// Start automatically after install.
 	err = svcAction(s, "start")
 	if err != nil {
-		log.Fatalf("Failed to start the service: %s", err)
+		log.Fatalf("service: starting: %s", err)
 	}
-	log.Printf("Service has been started")
+	log.Printf("service: started")
 
 	if detectFirstRun() {
 		log.Printf(`Almost ready!
 AdGuard Home is successfully installed and will automatically start on boot.
 There are a few more things that must be configured before you can use it.
-Click on the link below and follow the Installation Wizard steps to finish setup.`)
+Click on the link below and follow the Installation Wizard steps to finish setup.
+AdGuard Home is now available at the following addresses:`)
 		printHTTPAddresses(schemeHTTP)
 	}
 }
@@ -246,25 +275,24 @@ func handleServiceUninstallCommand(s service.Service) {
 		// as it will remove the symlink
 		_, err := runInitdCommand("disable")
 		if err != nil {
-			log.Fatal(err)
+			log.Fatalf("service: running init disable: %s", err)
 		}
 	}
 
-	err := svcAction(s, "uninstall")
-	if err != nil {
-		log.Fatal(err)
+	if err := svcAction(s, "uninstall"); err != nil {
+		log.Fatalf("service: executing action %q: %s", "uninstall", err)
 	}
 
 	if runtime.GOOS == "darwin" {
 		// Remove log files on cleanup and log errors.
-		err = os.Remove(launchdStdoutPath)
-		if err != nil && !os.IsNotExist(err) {
-			log.Printf("removing stdout file: %s", err)
+		err := os.Remove(launchdStdoutPath)
+		if err != nil && !errors.Is(err, os.ErrNotExist) {
+			log.Info("service: warning: removing stdout file: %s", err)
 		}
 
 		err = os.Remove(launchdStderrPath)
-		if err != nil && !os.IsNotExist(err) {
-			log.Printf("removing stderr file: %s", err)
+		if err != nil && !errors.Is(err, os.ErrNotExist) {
+			log.Info("service: warning: removing stderr file: %s", err)
 		}
 	}
 }
@@ -273,27 +301,40 @@ func handleServiceUninstallCommand(s service.Service) {
 func configureService(c *service.Config) {
 	c.Option = service.KeyValue{}
 
-	// OS X
+	// macOS
+
 	// Redefines the launchd config file template
 	// The purpose is to enable stdout/stderr redirect by default
 	c.Option["LaunchdConfig"] = launchdConfig
 	// This key is used to start the job as soon as it has been loaded. For daemons this means execution at boot time, for agents execution at login.
 	c.Option["RunAtLoad"] = true
 
-	// POSIX
-	// Redirect StdErr & StdOut to files.
-	c.Option["LogOutput"] = true
+	// POSIX / systemd
 
-	// Use modified service file templates.
+	// Redirect stderr and stdout to files.  Make sure we always restart.
+	c.Option["LogOutput"] = true
+	c.Option["Restart"] = "always"
+
+	// Start only once network is up on Linux/systemd.
+	if runtime.GOOS == "linux" {
+		c.Dependencies = []string{
+			"After=syslog.target network-online.target",
+		}
+	}
+
+	// Use the modified service file templates.
 	c.Option["SystemdScript"] = systemdScript
 	c.Option["SysvScript"] = sysvScript
 
-	// On OpenWrt we're using a different type of sysvScript.
+	// Use different scripts on OpenWrt and FreeBSD.
 	if aghos.IsOpenWrt() {
 		c.Option["SysvScript"] = openWrtScript
 	} else if runtime.GOOS == "freebsd" {
 		c.Option["SysvScript"] = freeBSDScript
 	}
+
+	c.Option["RunComScript"] = openBSDScript
+	c.Option["SvcInfo"] = fmt.Sprintf("%s %s", version.Full(), time.Now())
 }
 
 // runInitdCommand runs init.d service command
@@ -334,17 +375,26 @@ var launchdConfig = `<?xml version='1.0' encoding='UTF-8'?>
 </plist>
 `
 
-// Note: we should keep it in sync with the template from service_systemd_linux.go file
-// Add "After=" setting for systemd service file, because we must be started only after network is online
-// Set "RestartSec" to 10
+// systemdScript is an improved version of the systemd script originally from
+// the systemdScript constant in file service_systemd_linux.go in module
+// github.com/kardianos/service.  The following changes have been made:
+//
+// 1.  The RestartSec setting is set to a lower value of 10 to make sure we
+//     always restart quickly.
+//
+// 2.  The ExecStartPre setting is added to make sure that the log directory is
+//     always created to prevent the 209/STDOUT errors.
+//
 const systemdScript = `[Unit]
 Description={{.Description}}
 ConditionFileIsExecutable={{.Path|cmdEscape}}
-After=syslog.target network-online.target
+{{range $i, $dep := .Dependencies}}
+{{$dep}} {{end}}
 
 [Service]
 StartLimitInterval=5
 StartLimitBurst=10
+ExecStartPre=/bin/mkdir -p /var/log/
 ExecStart={{.Path|cmdEscape}}{{range .Arguments}} {{.|cmd}}{{end}}
 {{if .ChRoot}}RootDirectory={{.ChRoot|cmd}}{{end}}
 {{if .WorkingDirectory}}WorkingDirectory={{.WorkingDirectory|cmdEscape}}{{end}}
@@ -355,7 +405,9 @@ ExecStart={{.Path|cmdEscape}}{{range .Arguments}} {{.|cmd}}{{end}}
 StandardOutput=file:/var/log/{{.Name}}.out
 StandardError=file:/var/log/{{.Name}}.err
 {{- end}}
-Restart=always
+{{if gt .LimitNOFILE -1 }}LimitNOFILE={{.LimitNOFILE}}{{end}}
+{{if .Restart}}Restart={{.Restart}}{{end}}
+{{if .SuccessExitStatus}}SuccessExitStatus={{.SuccessExitStatus}}{{end}}
 RestartSec=10
 EnvironmentFile=-/etc/sysconfig/{{.Name}}
 
@@ -529,6 +581,20 @@ name="{{.Name}}"
 {{.Name}}_user="root"
 pidfile="/var/run/${name}.pid"
 command="/usr/sbin/daemon"
-command_args="-P ${pidfile} -f -r {{.WorkingDirectory}}/{{.Name}}"
+command_args="-p ${pidfile} -f -r {{.WorkingDirectory}}/{{.Name}}"
 run_rc_command "$1"
 `
+
+const openBSDScript = `#!/bin/sh
+#
+# $OpenBSD: {{ .SvcInfo }}
+
+daemon="{{.Path}}"
+daemon_flags={{ .Arguments | args }}
+
+. /etc/rc.d/rc.subr
+
+rc_bg=YES
+
+rc_cmd $1
+`
