@@ -142,12 +142,17 @@ func (l *queryLog) searchFiles(
 
 		return entries, oldest, 0
 	}
-	defer r.Close()
+	defer func() {
+		derr := r.Close()
+		if derr != nil {
+			log.Error("querylog: closing file: %s", err)
+		}
+	}()
 
 	if params.olderThan.IsZero() {
 		err = r.SeekStart()
 	} else {
-		err = r.SeekTS(params.olderThan.UnixNano())
+		err = r.seekTS(params.olderThan.UnixNano())
 		if err == nil {
 			// Read to the next record, because we only need the one
 			// that goes after it.
