@@ -4,7 +4,6 @@ import (
 	"bytes"
 	"encoding/binary"
 	"encoding/gob"
-	"errors"
 	"fmt"
 	"net"
 	"os"
@@ -12,7 +11,7 @@ import (
 	"sync"
 	"time"
 
-	"github.com/AdguardTeam/AdGuardHome/internal/agherr"
+	"github.com/AdguardTeam/golibs/errors"
 	"github.com/AdguardTeam/golibs/log"
 	bolt "go.etcd.io/bbolt"
 )
@@ -27,11 +26,13 @@ const (
 
 // statsCtx - global context
 type statsCtx struct {
+	// mu protects unit.
+	mu *sync.Mutex
+	// current is the actual statistics collection result.
+	current *unit
+
 	db   *bolt.DB
 	conf *Config
-
-	unit     *unit      // the current unit
-	unitLock sync.Mutex // protect 'unit'
 }
 
 // data for 1 time unit
@@ -67,7 +68,9 @@ type unitDB struct {
 }
 
 func createObject(conf Config) (s *statsCtx, err error) {
-	s = &statsCtx{}
+	s = &statsCtx{
+		mu: &sync.Mutex{},
+	}
 	if !checkInterval(conf.LimitDays) {
 		conf.LimitDays = 1
 	}
@@ -91,29 +94,7 @@ func createObject(conf Config) (s *statsCtx, err error) {
 		firstID := id - s.conf.limit - 1
 		unitDel := 0
 
-		// TODO(a.garipov): See if this is actually necessary.  Looks
-		// like a rather bizarre solution.
-		errStop := agherr.Error("stop iteration")
-		forEachBkt := func(name []byte, _ *bolt.Bucket) (cberr error) {
-			nameID := uint32(btoi(name))
-			if nameID < firstID {
-				cberr = tx.DeleteBucket(name)
-				if cberr != nil {
-					log.Debug("stats: tx.DeleteBucket: %s", cberr)
-
-					return nil
-				}
-
-				log.Debug("stats: deleted unit %d", nameID)
-				unitDel++
-
-				return nil
-			}
-
-			return errStop
-		}
-
-		err = tx.ForEach(forEachBkt)
+		err = tx.ForEach(newBucketWalker(tx, &unitDel, firstID))
 		if err != nil && !errors.Is(err, errStop) {
 			log.Debug("stats: deleting units: %s", err)
 		}
@@ -135,20 +116,53 @@ func createObject(conf Config) (s *statsCtx, err error) {
 	if udb != nil {
 		deserialize(&u, udb)
 	}
-	s.unit = &u
+	s.current = &u
 
 	log.Debug("stats: initialized")
 
 	return s, nil
 }
 
+// TODO(a.garipov): See if this is actually necessary.  Looks like a rather
+// bizarre solution.
+const errStop errors.Error = "stop iteration"
+
+// newBucketWalker returns a new bucket walker that deletes old units.  The
+// integer that unitDelPtr points to is incremented for every successful
+// deletion.  If the bucket isn't deleted, f returns errStop.
+func newBucketWalker(
+	tx *bolt.Tx,
+	unitDelPtr *int,
+	firstID uint32,
+) (f func(name []byte, b *bolt.Bucket) (err error)) {
+	return func(name []byte, _ *bolt.Bucket) (err error) {
+		nameID, ok := unitNameToID(name)
+		if !ok || nameID < firstID {
+			err = tx.DeleteBucket(name)
+			if err != nil {
+				log.Debug("stats: tx.DeleteBucket: %s", err)
+
+				return nil
+			}
+
+			log.Debug("stats: deleted unit %d (name %x)", nameID, name)
+
+			*unitDelPtr++
+
+			return nil
+		}
+
+		return errStop
+	}
+}
+
 func (s *statsCtx) Start() {
 	s.initWeb()
 	go s.periodicFlush()
 }
 
 func checkInterval(days uint32) bool {
-	return days == 1 || days == 7 || days == 30 || days == 90
+	return days == 0 || days == 1 || days == 7 || days == 30 || days == 90
 }
 
 func (s *statsCtx) dbOpen() bool {
@@ -168,11 +182,13 @@ func (s *statsCtx) dbOpen() bool {
 
 // Atomically swap the currently active unit with a new value
 // Return old value
-func (s *statsCtx) swapUnit(new *unit) *unit {
-	s.unitLock.Lock()
-	u := s.unit
-	s.unit = new
-	s.unitLock.Unlock()
+func (s *statsCtx) swapUnit(new *unit) (u *unit) {
+	s.mu.Lock()
+	defer s.mu.Unlock()
+
+	u = s.current
+	s.current = new
+
 	return u
 }
 
@@ -216,21 +232,35 @@ func (s *statsCtx) commitTxn(tx *bolt.Tx) {
 	log.Tracef("tx.Commit")
 }
 
-// Get unit name
-func unitName(id uint32) []byte {
-	return itob(uint64(id))
+// bucketNameLen is the length of a bucket, a 64-bit unsigned integer.
+//
+// TODO(a.garipov): Find out why a 64-bit integer is used when IDs seem to
+// always be 32 bits.
+const bucketNameLen = 8
+
+// idToUnitName converts a numerical ID into a database unit name.
+func idToUnitName(id uint32) (name []byte) {
+	name = make([]byte, bucketNameLen)
+	binary.BigEndian.PutUint64(name, uint64(id))
+
+	return name
 }
 
-// Convert integer to 8-byte array (big endian)
-func itob(v uint64) []byte {
-	b := make([]byte, 8)
-	binary.BigEndian.PutUint64(b, v)
-	return b
+// unitNameToID converts a database unit name into a numerical ID.  ok is false
+// if name is not a valid database unit name.
+func unitNameToID(name []byte) (id uint32, ok bool) {
+	if len(name) < bucketNameLen {
+		return 0, false
+	}
+
+	return uint32(binary.BigEndian.Uint64(name)), true
 }
 
-// Convert 8-byte array (big endian) to integer
-func btoi(b []byte) uint64 {
-	return binary.BigEndian.Uint64(b)
+func (s *statsCtx) ongoing() (u *unit) {
+	s.mu.Lock()
+	defer s.mu.Unlock()
+
+	return s.current
 }
 
 // Flush the current unit to DB and delete an old unit when a new hour is started
@@ -243,16 +273,13 @@ func btoi(b []byte) uint64 {
 // . unlock DB
 func (s *statsCtx) periodicFlush() {
 	for {
-		s.unitLock.Lock()
-		ptr := s.unit
-		s.unitLock.Unlock()
-
+		ptr := s.ongoing()
 		if ptr == nil {
 			break
 		}
 
 		id := s.conf.UnitID()
-		if ptr.id == id {
+		if ptr.id == id || s.conf.limit == 0 {
 			time.Sleep(time.Second)
 
 			continue
@@ -283,7 +310,7 @@ func (s *statsCtx) periodicFlush() {
 
 // Delete unit's data from file
 func (s *statsCtx) deleteUnit(tx *bolt.Tx, id uint32) bool {
-	err := tx.DeleteBucket(unitName(id))
+	err := tx.DeleteBucket(idToUnitName(id))
 	if err != nil {
 		log.Tracef("stats: bolt DeleteBucket: %s", err)
 
@@ -358,7 +385,7 @@ func deserialize(u *unit, udb *unitDB) {
 func (s *statsCtx) flushUnitToDB(tx *bolt.Tx, id uint32, udb *unitDB) bool {
 	log.Tracef("Flushing unit %d", id)
 
-	bkt, err := tx.CreateBucketIfNotExists(unitName(id))
+	bkt, err := tx.CreateBucketIfNotExists(idToUnitName(id))
 	if err != nil {
 		log.Error("tx.CreateBucketIfNotExists: %s", err)
 		return false
@@ -382,7 +409,7 @@ func (s *statsCtx) flushUnitToDB(tx *bolt.Tx, id uint32, udb *unitDB) bool {
 }
 
 func (s *statsCtx) loadUnitFromDB(tx *bolt.Tx, id uint32) *unitDB {
-	bkt := tx.Bucket(unitName(id))
+	bkt := tx.Bucket(idToUnitName(id))
 	if bkt == nil {
 		return nil
 	}
@@ -413,9 +440,11 @@ func convertTopSlice(a []countPair) []map[string]uint64 {
 }
 
 func (s *statsCtx) setLimit(limitDays int) {
-	conf := *s.conf
-	conf.limit = uint32(limitDays) * 24
-	s.conf = &conf
+	s.conf.limit = uint32(limitDays) * 24
+	if limitDays == 0 {
+		s.clear()
+	}
+
 	log.Debug("stats: set limit: %d", limitDays)
 }
 
@@ -472,23 +501,11 @@ func (s *statsCtx) clear() {
 	log.Debug("stats: cleared")
 }
 
-// Get Client IP address
-func (s *statsCtx) getClientIP(ip net.IP) (clientIP net.IP) {
-	if s.conf.AnonymizeClientIP && ip != nil {
-		const AnonymizeClientIP4Mask = 16
-		const AnonymizeClientIP6Mask = 112
-
-		if ip.To4() != nil {
-			return ip.Mask(net.CIDRMask(AnonymizeClientIP4Mask, 32))
-		}
-
-		return ip.Mask(net.CIDRMask(AnonymizeClientIP6Mask, 128))
+func (s *statsCtx) Update(e Entry) {
+	if s.conf.limit == 0 {
+		return
 	}
 
-	return ip
-}
-
-func (s *statsCtx) Update(e Entry) {
 	if e.Result == 0 ||
 		e.Result >= rLast ||
 		e.Domain == "" ||
@@ -498,14 +515,13 @@ func (s *statsCtx) Update(e Entry) {
 
 	clientID := e.Client
 	if ip := net.ParseIP(clientID); ip != nil {
-		ip = s.getClientIP(ip)
 		clientID = ip.String()
 	}
 
-	s.unitLock.Lock()
-	defer s.unitLock.Unlock()
+	s.mu.Lock()
+	defer s.mu.Unlock()
 
-	u := s.unit
+	u := s.current
 
 	u.nResult[e.Result]++
 
@@ -526,10 +542,8 @@ func (s *statsCtx) loadUnits(limit uint32) ([]*unitDB, uint32) {
 		return nil, 0
 	}
 
-	s.unitLock.Lock()
-	curUnit := serialize(s.unit)
-	curID := s.unit.id
-	s.unitLock.Unlock()
+	cur := s.ongoing()
+	curID := cur.id
 
 	// Per-hour units.
 	units := []*unitDB{}
@@ -545,7 +559,7 @@ func (s *statsCtx) loadUnits(limit uint32) ([]*unitDB, uint32) {
 
 	_ = tx.Rollback()
 
-	units = append(units, curUnit)
+	units = append(units, serialize(cur))
 
 	if len(units) != int(limit) {
 		log.Fatalf("len(units) != limit: %d %d", len(units), limit)
@@ -696,6 +710,10 @@ func (s *statsCtx) getData() (statsResponse, bool) {
 }
 
 func (s *statsCtx) GetTopClientsIP(maxCount uint) []net.IP {
+	if s.conf.limit == 0 {
+		return nil
+	}
+
 	units, _ := s.loadUnits(s.conf.limit)
 	if units == nil {
 		return nil
@@ -711,7 +729,10 @@ func (s *statsCtx) GetTopClientsIP(maxCount uint) []net.IP {
 	a := convertMapToSlice(m, int(maxCount))
 	d := []net.IP{}
 	for _, it := range a {
-		d = append(d, net.ParseIP(it.Name))
+		ip := net.ParseIP(it.Name)
+		if ip != nil {
+			d = append(d, ip)
+		}
 	}
 	return d
 }
